import falconjsonio.middleware, falconjsonio.schema

import falcon, falcon.testing
import json
import logging
import unittest


class NonJSONResource(object):
    def __init__(self):
        self.received = None

    def on_post(self, req, resp):
        self.received   = req.context['doc']
        resp.status     = falcon.HTTP_201
        resp.body       = self.received
        resp.set_header('Content-Type', req.content_type)

class SchemalessJSONResource(object):
    def __init__(self):
        self.received = None

    def on_post(self, req, resp):
        self.received           = req.context['doc']
        resp.status             = falcon.HTTP_201
        req.context['result']   = self.received
        resp.set_header('Content-Type', req.content_type)

class GoodResource(object):
    def __init__(self):
        self.received = None

    @falconjsonio.schema.request_schema({
        'type': 'object',
        'properties': {
            'email':    {'type': 'string'},
            'password': {'type': 'string'},
        },
        'required': ['email', 'password'],
    })
    @falconjsonio.schema.response_schema({
        'type': 'object',
        'properties': {
            'email': {'type': 'string'},
        },
        'required': ['email'],
    })
    def on_post(self, req, resp):
        self.received = req.context['doc']
        resp.status = falcon.HTTP_201
        req.context['result'] = {'email': req.context['doc']['email']}

    @falconjsonio.schema.response_schema({
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
        },
        'required': ['id'],
    })
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        req.context['result'] = {'id': 12345}

class GoodParentResource(object):
    def __init__(self):
        self.received = None

    def on_post(self, req, resp):
        self.received = req.context['doc']
        resp.status = falcon.HTTP_201
        req.context['result'] = {'email': req.context['doc']['email']}

@falconjsonio.schema.request_schema(schema={
    'type': 'object',
    'properties': {
        'email':    {'type': 'string'},
        'password': {'type': 'string'},
    },
    'required': ['email', 'password'],
}, method_name='on_post')
@falconjsonio.schema.response_schema(schema={
    'type': 'object',
    'properties': {
        'email': {'type': 'string'},
    },
    'required': ['email'],
}, method_name='on_post')
class GoodChildResource(GoodParentResource):
    pass

class BadParentResource(object):
    def on_post(self, req, resp):
        self.received = req.context['doc']
        resp.status = falcon.HTTP_201
        req.context['result'] = {'this': 'does not conform'}

@falconjsonio.schema.request_schema(schema={
    'type': 'object',
    'properties': {
        'email':    {'type': 'string'},
        'password': {'type': 'string'},
    },
    'required': ['email', 'password'],
}, method_name='on_post')
@falconjsonio.schema.response_schema(schema={
    'type': 'object',
    'properties': {
        'email': {'type': 'string'},
    },
    'required': ['email'],
}, method_name='on_post')
class BadChildResource(BadParentResource):
    pass

class BadResource(object):
    def __init__(self):
        self.received = None

    @falconjsonio.schema.request_schema({
        'type': 'object',
        'properties': {
            'email':    {'type': 'string'},
            'password': {'type': 'string'},
        },
        'required': ['email', 'password'],
    })
    @falconjsonio.schema.response_schema({
        'type': 'object',
        'properties': {
            'email': {'type': 'string'},
        },
        'required': ['email'],
    })
    def on_post(self, req, resp):
        self.received = req.context['doc']
        resp.status = falcon.HTTP_201
        req.context['result'] = {'this': 'does not conform'}

    @falconjsonio.schema.response_schema({
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
        },
        'required': ['id'],
    })
    def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        req.context['result'] = {'this': 'does not conform'}

class CollectingHandler(logging.Handler):
    def __init__(self):
        super(CollectingHandler, self).__init__()
        self.logs = []
    def emit(self, record):
        self.logs.append(record)

class IOTest(unittest.TestCase):
    def setUp(self):
        super(IOTest, self).setUp()

        self.logger = logging.getLogger('TestLogger')
        self.handler = CollectingHandler()
        self.logger.handlers = []
        self.logger.addHandler(self.handler)

        self.app = falcon.API(
            middleware=[
                falconjsonio.middleware.RequireJSON(),
                falconjsonio.middleware.JSONTranslator(self.logger),
            ],
        )

        self.non_json_resource          = NonJSONResource()
        self.schemaless_json_resource   = SchemalessJSONResource()
        self.good_resource              = GoodResource()
        self.bad_resource               = BadResource()
        self.good_child_resource        = GoodChildResource()
        self.bad_child_resource         = BadChildResource()
        self.app.add_route('/non_json_response',        self.non_json_resource)
        self.app.add_route('/schemaless_json_response', self.schemaless_json_resource)
        self.app.add_route('/good_response',            self.good_resource)
        self.app.add_route('/bad_response',             self.bad_resource)
        self.app.add_route('/good_child_response',      self.good_child_resource)
        self.app.add_route('/bad_child_response',       self.bad_child_resource)

        self.srmock = falcon.testing.StartResponseMock()

    def simulate_request(self, path, *args, **kwargs):
        env = falcon.testing.create_environ(path=path, **kwargs)
        return self.app(env, self.srmock)

    def test_no_accept(self):
        response, = self.simulate_request('/good_response', method='POST', body=json.dumps({}), headers={})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')
        self.assertEqual((json.loads(response.decode('utf-8')))['title'], 'Unsupported media type')

    def test_unsupported_accept(self):
        response = self.simulate_request('/good_response', method='POST', body=json.dumps({}), headers={'Accept': 'text/html'})
        self.assertEqual(len(response), 0) # Unsupported media type, so we can't respond with a body
        self.assertEqual(self.srmock.status, '406 Not Acceptable')

    def test_no_content_type(self):
        response = self.simulate_request('/good_response', method='POST', body=json.dumps({}), headers={'Accept': 'application/json'})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')

    def test_unsupported_content_type(self):
        response = self.simulate_request('/good_response', method='POST', body=json.dumps({}), headers={'Accept': 'application/json', 'Content-Type': 'text/html'})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')

    def test_non_json_endpoint(self):
        response = self.simulate_request('/non_json_response', method='POST', body='Hello this is some text', headers={'Accept': 'text/plain', 'Content-Type': 'text/plain'})
        self.assertEqual(self.srmock.status, '201 Created')
        self.assertEqual(response[0].decode('utf-8'), 'Hello this is some text')
        self.assertEqual(self.non_json_resource.received.decode('utf-8'), 'Hello this is some text')

    def test_schemaless_json_endpoint(self):
        response = self.simulate_request('/schemaless_json_response', method='POST', body=json.dumps({'hello': 'world'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.srmock.status, '201 Created')
        self.assertEqual(json.loads(response[0].decode('utf-8')), {'hello': 'world'})

    def test_post(self):
        response = self.simulate_request('/good_response', method='POST', body=json.dumps({'email': 'foo@example.com', 'password': 'hunter2'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.good_resource.received, {'email': 'foo@example.com', 'password': 'hunter2'})
        self.assertEqual(self.srmock.status, '201 Created')
        self.assertEqual(json.loads(response[0].decode('utf-8')), {'email': 'foo@example.com'})

    def test_nonexistent_method(self):
        response = self.simulate_request('/good_response', method='PUT', body=json.dumps({'email': 'foo@example.com', 'password': 'hunter2'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.srmock.status, '405 Method Not Allowed')

    def test_empty_post(self):
        response = self.simulate_request('/good_response', method='POST', headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.srmock.status, '400 Bad Request')

    def test_nonconforming_post_request(self):
        response = self.simulate_request('/good_response', method='POST', body=json.dumps({'email': 'foo@example.com'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.good_resource.received, None)
        self.assertEqual(self.srmock.status, '400 Bad Request')
        self.assertEqual((json.loads(response[0].decode('utf-8')))['title'], 'Invalid request body')

    def test_nonconforming_post_response(self):
        response, = self.simulate_request('/bad_response', method='POST', body=json.dumps({'email': 'foo@example.com', 'password': 'hunter2'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.bad_resource.received, {'email': 'foo@example.com', 'password': 'hunter2'})
        self.assertEqual(self.srmock.status, '500 Internal Server Error')
        # Programming error is logged
        self.assertEqual(
            [self.handler.logs[0].message],
            ["""Blocking proposed response from being sent from falconjsonio.test.BadResource.on_post to client as it does not match the defined schema: 'email' is a required property

Failed validating 'required' in schema:
    {'properties': {'email': {'type': 'string'}},
     'required': ['email'],
     'type': 'object'}

On instance:
    {'this': 'does not conform'}"""]
        )

        self.assertEqual(json.loads(response.decode('utf-8')), {'title': 'Internal Server Error', 'description': 'Undisclosed'})

###
    def test_no_accept_inherited(self):
        response, = self.simulate_request('/good_child_response', method='POST', body=json.dumps({}), headers={})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')
        self.assertEqual((json.loads(response.decode('utf-8')))['title'], 'Unsupported media type')

    def test_unsupported_accept_inherited(self):
        response = self.simulate_request('/good_child_response', method='POST', body=json.dumps({}), headers={'Accept': 'text/html'})
        self.assertEqual(len(response), 0) # Unsupported media type, so we can't respond with a body
        self.assertEqual(self.srmock.status, '406 Not Acceptable')

    def test_no_content_type_inherited(self):
        response = self.simulate_request('/good_child_response', method='POST', body=json.dumps({}), headers={'Accept': 'application/json'})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')

    def test_unsupported_content_type_inherited(self):
        response = self.simulate_request('/good_child_response', method='POST', body=json.dumps({}), headers={'Accept': 'application/json', 'Content-Type': 'text/html'})
        self.assertEqual(self.srmock.status, '415 Unsupported Media Type')

    def test_post_inherited(self):
        response = self.simulate_request('/good_child_response', method='POST', body=json.dumps({'email': 'foo@example.com', 'password': 'hunter2'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.good_child_resource.received, {'email': 'foo@example.com', 'password': 'hunter2'})
        self.assertEqual(self.srmock.status, '201 Created')
        self.assertEqual(json.loads(response[0].decode('utf-8')), {'email': 'foo@example.com'})

    def test_nonconforming_post_request_inherited(self):
        response = self.simulate_request('/good_child_response', method='POST', body=json.dumps({'email': 'foo@example.com'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.good_child_resource.received, None)
        self.assertEqual(self.srmock.status, '400 Bad Request')
        self.assertEqual((json.loads(response[0].decode('utf-8')))['title'], 'Invalid request body')

    def test_nonconforming_post_response_inherited(self):
        response, = self.simulate_request('/bad_child_response', method='POST', body=json.dumps({'email': 'foo@example.com', 'password': 'hunter2'}), headers={'Accept': 'application/json', 'Content-Type': 'application/json'})
        self.assertEqual(self.bad_child_resource.received, {'email': 'foo@example.com', 'password': 'hunter2'})
        self.assertEqual(self.srmock.status, '500 Internal Server Error')
        # Programming error is logged
        self.assertEqual(
            [self.handler.logs[0].message],
            ["""Blocking proposed response from being sent from falconjsonio.test.BadChildResource.on_post to client as it does not match the defined schema: 'email' is a required property

Failed validating 'required' in schema:
    {'properties': {'email': {'type': 'string'}},
     'required': ['email'],
     'type': 'object'}

On instance:
    {'this': 'does not conform'}"""]
        )

        self.assertEqual(json.loads(response.decode('utf-8')), {'title': 'Internal Server Error', 'description': 'Undisclosed'})

    def test_invalid_class_request_schema_call(self):
        with self.assertRaises(falconjsonio.schema.SchemaDecoratorError):
            @falconjsonio.schema.request_schema({})
            class Foo(object):
                pass

        with self.assertRaises(falconjsonio.schema.SchemaDecoratorError):
            @falconjsonio.schema.response_schema({})
            class Foo(object):
                pass



import inspect

class SchemaDecoratorError(Exception): pass

class request_schema(object):
    """
    Decorator to specify the JSON schema required for a request.
    """
    def __init__(self, schema, method_name=None):
        self.schema         = schema
        self.method_name    = method_name

    def __call__(self, klass_or_func):
        if inspect.isclass(klass_or_func):
            if not hasattr(klass_or_func, '__request_schemas__'):
                klass_or_func.__request_schemas__ = {}
            if self.method_name is None:
                raise SchemaDecoratorError("Parameter 'method_name' must be supplied when applying request_schema decorator to a class")
            klass_or_func.__request_schemas__[self.method_name] = self.schema
        else:
            klass_or_func.__request_schema__ = self.schema
        return klass_or_func

class response_schema(object):
    """
    Decorator to specify the JSON schema a response should conform to.
    """
    def __init__(self, schema, method_name = None):
        self.schema         = schema
        self.method_name    = method_name

    def __call__(self, klass_or_func):
        if inspect.isclass(klass_or_func):
            if not hasattr(klass_or_func, '__response_schemas__'):
                klass_or_func.__response_schemas__ = {}
            if self.method_name is None:
                raise SchemaDecoratorError("Parameter 'method_name' must be supplied when applying response_schema decorator to a class")
            klass_or_func.__response_schemas__[self.method_name] = self.schema
        else:
            klass_or_func.__response_schema__ = self.schema
        return klass_or_func



import falcon
import json
import jsonschema
import logging


def _get_request_schema(req, resource):
    if resource is None or req.method not in ['POST', 'PUT', 'PATCH']:
        return None

    # First try to get schema from method itself
    schema = getattr(
        getattr(resource, {'POST': 'on_post', 'PUT': 'on_put', 'PATCH': 'on_patch'}[req.method], None),
        '__request_schema__',
        None
    # Otherwise, fall back to schema defined directly in class
    ) or getattr(resource, '__request_schemas__', {}).get(
        {'POST': 'on_post', 'PUT': 'on_put', 'PATCH': 'on_patch'}[req.method]
    )
    return schema

def _get_response_schema(resource, req):
    try:
        method_name = {'POST': 'on_post', 'PUT': 'on_put', 'PATCH': 'on_patch', 'GET': 'on_get', 'DELETE': 'on_delete'}[req.method]
    except KeyError:
        return

    # First try to get schema from method itself
    return getattr(
        getattr(resource, method_name, None),
        '__response_schema__',
        None
    # Otherwise, fall back to schema defined directly in class
    ) or getattr(resource, '__response_schemas__', {}).get(method_name)


class _null_handler(logging.Handler):
    def emit(self, record):
        pass

class RequireJSON(object):
    def process_resource(self, req, resp, resource):
        if _get_response_schema(resource, req) and not req.client_accepts_json:
            raise falcon.HTTPNotAcceptable('This API supports only JSON-encoded responses')
        if req.method in ('POST', 'PUT', 'PATCH'):
            if _get_request_schema(req, resource) is not None:
                if req.content_type is None or 'application/json' not in req.content_type:
                    raise falcon.HTTPUnsupportedMediaType('This API supports only JSON-encoded requests')

class JSONTranslator(object):
    def __init__(self, logger=None):
        if logger is None:
            # Default to no logging if no logger provided
            logger = logging.getLogger(__name__)
            logger.addHandler(_null_handler())
        self.logger = logger

    def process_resource(self, req, resp, resource):
        if resource is None or req.method not in ['POST', 'PUT', 'PATCH']:
            return

        body = req.stream.read()
        if not body:
            raise falcon.HTTPBadRequest(
                'Empty request body',
                'A valid JSON document is required'
            )

        schema = _get_request_schema(req, resource)
        try:
            req.context['doc'] = json.loads(body.decode('utf-8'))
        except (ValueError, UnicodeDecodeError) as error:
            if schema is not None:
                raise falcon.HTTPBadRequest(
                    'Malformed JSON',
                    'Could not decode the request body.  The JSON was incorrect or not encoded as UTF-8'
                )
            req.context['doc'] = body

        if schema is not None:
            try:
                jsonschema.validate(req.context['doc'], schema)
            except jsonschema.exceptions.ValidationError as error:
                raise falcon.HTTPBadRequest(
                    'Invalid request body',
                    json.dumps({'error': str(error)})
                )

    def process_response(self, req, resp, resource):
        if 'result' not in req.context:
            return

        resp.body = json.dumps(req.context['result'])

        schema = _get_response_schema(resource, req)
        if schema is None:
            return

        try:
            jsonschema.validate(req.context['result'], schema)
        except jsonschema.exceptions.ValidationError as error:
            method_name = {'POST': 'on_post', 'PUT': 'on_put', 'PATCH': 'on_patch', 'GET': 'on_get', 'DELETE': 'on_delete'}[req.method]
            self.logger.error('Blocking proposed response from being sent from {0}.{1}.{2} to client as it does not match the defined schema: {3}'.format(resource.__module__, resource.__class__.__name__, method_name, str(error)))
            raise falcon.HTTPInternalServerError('Internal Server Error', 'Undisclosed')



from setuptools import setup
from pip.req import parse_requirements
from pip.download import PipSession

version = '0.1.10'

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = open('README.md').read()

setup(name='falconjsonio',
      version=version,
      description='JSON-Schema input and output for Falcon',
      long_description=long_description,
      url='https://bitbucket.org/garymonson/falcon-json-io',
      author='Gary Monson',
      author_email='gary.monson@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
      ],
      keywords='json schema falcon',
      packages=['falconjsonio'],
      install_requires=[str(ir.req) for ir in parse_requirements('requirements.txt', session=PipSession())],
      zip_safe=False)
